﻿using System;
using System.Linq;
using System.Collections.Generic;

namespace ConsoleApplication1
{
    class Person
    {
        public string Name { get; set; }
        public int Age { get; set; }
        public string Address { get; set; }

        public override string ToString()
        {
            return string.Format("{0}: {1} in {2}", Name, Age, Address);
        }
    }

    class MainLanguage
    {
        public string Name { get; set; }
        public string Language { get; set; }
    }

    class Program
    {
        static void Main(string[] args)
        {
            List<Person> people = new List<Person>
            {
                new Person { Name = "Tom", Age = 63, Address = "Korea" },
                new Person { Name = "Winnie", Age = 40, Address = "Tibet" },
                new Person { Name = "Anders", Age = 47, Address = "Sudan" },
                new Person { Name = "Hans", Age = 25, Address = "Tibet" },
                new Person { Name = "Eureka", Age = 32, Address = "Sudan" },
                new Person { Name = "Hawk", Age = 15, Address = "Korea" },
            };

            List<MainLanguage> languages = new List<MainLanguage>
            {
                new MainLanguage { Name = "Anders", Language = "Delphi" },
                new MainLanguage { Name = "Anders", Language = "C#" },
                new MainLanguage { Name = "Tom", Language = "Borland C++" },
                new MainLanguage { Name = "Hans", Language = "Visual C++" },
                new MainLanguage { Name = "Winnie", Language = "R" },
            };

            // 가장 간단한 유형의 LINQ 쿼리
            IEnumerable<Person> all = from person in people select person;

            foreach (var item in all)
            {
                Console.WriteLine(item);
            }
            Console.WriteLine("-----------------------------------");

            // LINQ 쿼리를 확장 메소드로 대체
            IEnumerable<Person> all2 = people.Select((elem) => elem);
            foreach (var item in all2)
            {
                Console.WriteLine(item);
            }
            Console.WriteLine("-----------------------------------");

            // LINQ 쿼리를 IEnumerable<T>를 반환하는 메소드로 대체
            SelectFunc(people);

            // 반환 타입을 string 목록으로 변환
            var nameList = from person in people
                           select person.Name;

            foreach (var item in nameList)
            {
                Console.WriteLine(item);
            }
            Console.WriteLine("-----------------------------------");

            // LINQ 쿼리를 확장 메소드로 표현
            var nameList2 = people.Select((elem) => elem.Name);
            foreach (var item in nameList2)
            {
                Console.WriteLine(item);
            }
            Console.WriteLine("-----------------------------------");

            // 반환 타입을 익명 타입의 목록으로 변환
            var dateList = from person in people
                           select new { Name = person.Name, Year = DateTime.Now.AddYears(-person.Age).Year };
            foreach (var item in dateList)
            {
                Console.WriteLine(string.Format("{0} - {1}", item.Name, item.Year));
            }
            Console.WriteLine("-----------------------------------");

            // 익명 타입을 사용하는 LINQ 쿼리를 확장 메소드로 표현
            var dateList2 = people.Select((elem) => new { Name = elem.Name, Year = DateTime.Now.AddYears(-elem.Age).Year });
            foreach (var item in dateList2)
            {
                Console.WriteLine(string.Format("{0} - {1}", item.Name, item.Year));
            }
            Console.WriteLine("-----------------------------------");

            // where 구문 사용
            var ageOver30 = from person in people
                            where person.Age > 30
                            select person;

            foreach (var item in ageOver30)
            {
                Console.WriteLine(item);
            }
            Console.WriteLine("-----------------------------------");

            // where 구문에 자유롭게 메소드 호출 가능
            var endWithS = from person in people
                           where person.Name.EndsWith("s")
                           select person;

            foreach (var item in endWithS)
            {
                Console.WriteLine(item);
            }
            Console.WriteLine("-----------------------------------");

            // orderby 를 이용한 오름차순 정렬
            var ageSort = from person in people
                          orderby person.Age // 나이순으로 정렬
                          select person;

            foreach (var item in ageSort)
            {
                Console.WriteLine(item);
            }
            Console.WriteLine("-----------------------------------");

            // 내림차순 정렬
            var ageSortDesc = from person in people
                              orderby person.Age descending
                              select person;

            foreach (var item in ageSortDesc)
            {
                Console.WriteLine(item);
            }
            Console.WriteLine("-----------------------------------");

            // group ... by ...
            var addrGroup = from person in people
                            group person by person.Address;

            foreach (var itemGroup in addrGroup)
            {
                Console.WriteLine(string.Format("[{0}]", itemGroup.Key));
                foreach (var item in itemGroup)
                {
                    Console.WriteLine(item);
                }
                Console.WriteLine();
            }
            Console.WriteLine("-----------------------------------");

            // group ... by ... 에 형변환 적용
            var nameAgeList = from person in people
                              group new { Name = person.Name, Age = person.Age } by person.Address;

            foreach (var itemGroup in nameAgeList)
            {
                Console.WriteLine(string.Format("[{0}]", itemGroup.Key));
                foreach (var item in itemGroup)
                {
                    Console.WriteLine(item);
                }
                Console.WriteLine();
            }
            Console.WriteLine("-----------------------------------");

            // inner join
            var nameToLangList = from person in people
                                 join language in languages on person.Name equals language.Name
                                 select new { Name = person.Name, Age = person.Age, Language = language.Language };

            foreach (var item in nameToLangList)
            {
                Console.WriteLine(item);
            }
            Console.WriteLine("-----------------------------------");

            // outer join
            var nameToLangAllList = from person in people
                                    join language in languages on person.Name equals language.Name into lang
                                    from language in lang.DefaultIfEmpty(new MainLanguage())
                                    select new { Name = person.Name, Age = person.Age, Language = language.Language };

            foreach (var item in nameToLangAllList)
            {
                Console.WriteLine(item);
            }
            Console.WriteLine("-----------------------------------");
        }

        static IEnumerable<Person> SelectFunc(List<Person> people)
        {
            foreach (var item in people)
            {
                yield return item;
            }
        }
    }
}
